"""
Loader code for some datasets.
"""
from .base import WindowsDataset, BaseDataset, BaseConcatDataset
from .moabb import MOABBDataset
from .tuh import TUH, TUHAbnormal
from .sleep_physionet import SleepPhysionet
